@@ -83,6 +83,11 @@ static const struct value_string stream_cli_state_names[] = {
 #define OSMO_STREAM_CLI_F_RECONF	(1 << 0)
 #define OSMO_STREAM_CLI_F_NODELAY	(1 << 1)
 
+/* Mark whether the object is currently in a user callback. */
+#define IN_CB_MASK_CONNECT_CB		(1 << 0)
+#define IN_CB_MASK_DISCONNECT_CB	(1 << 1)
+#define IN_CB_MASK_READ_CB		(1 << 2)
+
 struct osmo_stream_cli {
 	char *name;
 	char sockname[OSMO_SOCK_NAME_MAXLEN];
@@ -91,7 +96,9 @@ struct osmo_stream_cli {
 		struct osmo_fd			ofd;
 		struct osmo_io_fd		*iofd;
 	};
-	struct llist_head		tx_queue;
+	struct llist_head		tx_queue; /* osmo_ofd mode (only): Queue of msgbs */
+	unsigned int			tx_queue_count; /* osmo_ofd mode (only): Current amount of msgbs queued */
+	unsigned int			tx_queue_max_length; /* Max amount of msgbs which can be enqueued */
 	struct osmo_timer_list		timer;
 	enum osmo_stream_cli_state	state;
 	char				*addr[OSMO_STREAM_MAX_ADDRS];
@@ -102,7 +109,9 @@ struct osmo_stream_cli {
 	uint16_t			local_port;
 	int				sk_domain;
 	int				sk_type;
+	int				sk_prio; /* socket priority, SO_PRIORITY, default=0=unset */
 	uint16_t			proto;
+	uint8_t				ip_dscp; /* IP Differentiated services, 0..63, default=0=unset */
 	osmo_stream_cli_connect_cb_t	connect_cb;
 	osmo_stream_cli_disconnect_cb_t	disconnect_cb;
 	osmo_stream_cli_read_cb_t	read_cb;
@@ -112,31 +121,136 @@ struct osmo_stream_cli {
 	int				flags;
 	int				reconnect_timeout;
 	struct osmo_sock_init2_multiaddr_pars ma_pars;
+	uint8_t				in_cb_mask; /* IN_CB_MASK_* */
+	bool				delay_free;
 };
 
-void osmo_stream_cli_close(struct osmo_stream_cli *cli);
-
 /*! \addtogroup stream_cli
  *  @{
  */
 
+/* return true if freed */
+static inline bool free_delayed_if_needed(struct osmo_stream_cli *cli)
+{
+	/* Nobody requested delayed free, skip */
+	if (!cli->delay_free)
+		return false;
+	/* Check for other callbacks active in case we were e.g. in:
+	* read_cb() -> [user] -> osmo_steam_client_close() -> disconnect_cb() --> [user] --> osmo_stream_client_destroy()
+	* or:
+	* connect_cb() -> [user] -> osmo_steam_client_close() -> disconnect_cb() --> [user] --> osmo_stream_client_destroy()
+	*/
+	if (cli->in_cb_mask != 0)
+		return false;
+
+	LOGSCLI(cli, LOGL_DEBUG, "free(delayed)\n");
+	talloc_free(cli);
+	return true;
+}
+
+static void stream_cli_close_iofd(struct osmo_stream_cli *cli)
+{
+	if (!cli->iofd)
+		return;
+
+	osmo_iofd_free(cli->iofd);
+	cli->iofd = NULL;
+}
+
+static void stream_cli_close_ofd(struct osmo_stream_cli *cli)
+{
+	if (cli->ofd.fd == -1)
+		return;
+	osmo_fd_unregister(&cli->ofd);
+	close(cli->ofd.fd);
+	cli->ofd.fd = -1;
+}
+
+/*! Close an Osmocom Stream Client.
+ *  \param[in] cli Osmocom Stream Client to be closed
+ *  \return true if stream was freed due to disconnect_cb, false otherwise
+ *  We unregister the socket fd from the osmocom select() loop
+ *  abstraction and close the socket */
+static bool stream_cli_close(struct osmo_stream_cli *cli)
+{
+	int old_state = cli->state;
+	LOGSCLI(cli, LOGL_DEBUG, "close()\n");
+
+	/* This guards against reentrant close through disconnect_cb(): */
+	if (cli->state == STREAM_CLI_STATE_CLOSED)
+		return false;
+	if (cli->state == STREAM_CLI_STATE_WAIT_RECONNECT) {
+		osmo_timer_del(&cli->timer);
+		cli->state = STREAM_CLI_STATE_CLOSED;
+		return false;
+	}
+
+
+	switch (cli->mode) {
+	case OSMO_STREAM_MODE_OSMO_FD:
+		stream_cli_close_ofd(cli);
+		break;
+	case OSMO_STREAM_MODE_OSMO_IO:
+		stream_cli_close_iofd(cli);
+		break;
+	default:
+		OSMO_ASSERT(false);
+	}
+
+	cli->state = STREAM_CLI_STATE_CLOSED;
+
+	/* If conn was established, notify the disconnection to the user:
+	 * Also, if reconnect is disabled by user, notify the user that connect() failed: */
+	if (old_state == STREAM_CLI_STATE_CONNECTED ||
+	    (old_state == STREAM_CLI_STATE_CONNECTING && cli->reconnect_timeout < 0)) {
+		cli->in_cb_mask |= IN_CB_MASK_DISCONNECT_CB;
+		if (cli->disconnect_cb)
+			cli->disconnect_cb(cli);
+		cli->in_cb_mask &= ~IN_CB_MASK_DISCONNECT_CB;
+		return free_delayed_if_needed(cli);
+	}
+	return false;
+}
+
+/*! Close an Osmocom Stream Client.
+ *  \param[in] cli Osmocom Stream Client to be closed
+ *  We unregister the socket fd from the osmocom select() loop
+ *  abstraction and close the socket */
+void osmo_stream_cli_close(struct osmo_stream_cli *cli)
+{
+	stream_cli_close(cli);
+}
+
+/*! Re-connect an Osmocom Stream Client.
+ *  If re-connection is enabled for this client
+ *  (which is the case unless negative timeout was explicitly set via osmo_stream_cli_set_reconnect_timeout() call),
+ *  we close any existing connection (if any) and schedule a re-connect timer */
+static bool stream_cli_reconnect(struct osmo_stream_cli *cli)
+{
+	bool freed = stream_cli_close(cli);
+
+	if (freed)
+		return true;
+
+	if (cli->reconnect_timeout < 0) {
+		LOGSCLI(cli, LOGL_INFO, "not reconnecting, disabled\n");
+		return false;
+	}
+
+	cli->state = STREAM_CLI_STATE_WAIT_RECONNECT;
+	LOGSCLI(cli, LOGL_INFO, "retrying reconnect in %d seconds...\n",
+		cli->reconnect_timeout);
+	osmo_timer_schedule(&cli->timer, cli->reconnect_timeout, 0);
+	return false;
+}
+
 /*! Re-connect an Osmocom Stream Client.
  *  If re-connection is enabled for this client
  *  (which is the case unless negative timeout was explicitly set via osmo_stream_cli_set_reconnect_timeout() call),
  *  we close any existing connection (if any) and schedule a re-connect timer */
 void osmo_stream_cli_reconnect(struct osmo_stream_cli *cli)
 {
-	osmo_stream_cli_close(cli);
-
-	if (cli->reconnect_timeout < 0) {
-		LOGSCLI(cli, LOGL_INFO, "not reconnecting, disabled\n");
-		return;
-	}
-
-	cli->state = STREAM_CLI_STATE_WAIT_RECONNECT;
-	LOGSCLI(cli, LOGL_INFO, "retrying reconnect in %d seconds...\n",
-		cli->reconnect_timeout);
-	osmo_timer_schedule(&cli->timer, cli->reconnect_timeout, 0);
+	stream_cli_reconnect(cli);
 }
 
 /*! Check if Osmocom Stream Client is in connected state.
@@ -159,61 +273,6 @@ static bool stream_cli_is_opened(const struct osmo_stream_cli *cli)
 	       cli->state == STREAM_CLI_STATE_CONNECTED;
 }
 
-static void osmo_stream_cli_close_iofd(struct osmo_stream_cli *cli)
-{
-	if (!cli->iofd)
-		return;
-
-	osmo_iofd_free(cli->iofd);
-	cli->iofd = NULL;
-}
-
-static void osmo_stream_cli_close_ofd(struct osmo_stream_cli *cli)
-{
-	if (cli->ofd.fd == -1)
-		return;
-	osmo_fd_unregister(&cli->ofd);
-	close(cli->ofd.fd);
-	cli->ofd.fd = -1;
-}
-
-/*! Close an Osmocom Stream Client.
- *  \param[in] cli Osmocom Stream Client to be closed
- *  We unregister the socket fd from the osmocom select() loop
- *  abstraction and close the socket */
-void osmo_stream_cli_close(struct osmo_stream_cli *cli)
-{
-	int old_state = cli->state;
-
-	if (cli->state == STREAM_CLI_STATE_CLOSED)
-		return;
-	if (cli->state == STREAM_CLI_STATE_WAIT_RECONNECT) {
-		osmo_timer_del(&cli->timer);
-		cli->state = STREAM_CLI_STATE_CLOSED;
-		return;
-	}
-
-
-	switch (cli->mode) {
-	case OSMO_STREAM_MODE_OSMO_FD:
-		osmo_stream_cli_close_ofd(cli);
-		break;
-	case OSMO_STREAM_MODE_OSMO_IO:
-		osmo_stream_cli_close_iofd(cli);
-		break;
-	default:
-		OSMO_ASSERT(false);
-	}
-
-	cli->state = STREAM_CLI_STATE_CLOSED;
-
-	if (old_state == STREAM_CLI_STATE_CONNECTED) {
-		LOGSCLI(cli, LOGL_DEBUG, "connection closed\n");
-		if (cli->disconnect_cb)
-			cli->disconnect_cb(cli);
-	}
-}
-
 /*! Retrieve file descriptor of the stream client socket.
  *  \param[in] cli Stream Client of which we want to obtain the file descriptor
  *  \returns File descriptor or negative in case of error */
@@ -244,15 +303,19 @@ osmo_stream_cli_get_iofd(const struct osmo_stream_cli *cli)
 	return cli->iofd;
 }
 
-static void osmo_stream_cli_read(struct osmo_stream_cli *cli)
+/* Return true if read_cb caused a delayed_free, hence cli not available anymore. */
+static bool stream_cli_read(struct osmo_stream_cli *cli)
 {
 	LOGSCLI(cli, LOGL_DEBUG, "message received\n");
 
+	cli->in_cb_mask |= IN_CB_MASK_READ_CB;
 	if (cli->read_cb)
 		cli->read_cb(cli);
+	cli->in_cb_mask &= ~IN_CB_MASK_READ_CB;
+	return free_delayed_if_needed(cli);
 }
 
-static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
+static int stream_cli_write(struct osmo_stream_cli *cli)
 {
 #ifdef HAVE_LIBSCTP
 	struct sctp_sndrcvinfo sinfo;
@@ -260,12 +323,11 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 	struct msgb *msg;
 	int ret;
 
-	if (llist_empty(&cli->tx_queue)) {
+	msg = msgb_dequeue_count(&cli->tx_queue, &cli->tx_queue_count);
+	if (!msg) { /* done, tx_queue empty */
 		osmo_fd_write_disable(&cli->ofd);
 		return 0;
 	}
-	msg = llist_first_entry(&cli->tx_queue, struct msgb, list);
-	llist_del(&msg->list);
 
 	if (!osmo_stream_cli_is_connected(cli)) {
 		LOGSCLI(cli, LOGL_ERROR, "send: not connected, dropping data!\n");
@@ -306,6 +368,7 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 		/* Update msgb and re-add it at the start of the queue: */
 		msgb_pull(msg, ret);
 		llist_add(&msg->list, &cli->tx_queue);
+		cli->tx_queue_count++;
 		return 0;
 	}
 
@@ -315,10 +378,11 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 		if (err == EAGAIN) {
 			/* Re-add at the start of the queue to re-attempt: */
 			llist_add(&msg->list, &cli->tx_queue);
+			cli->tx_queue_count++;
 			return 0;
 		}
 		msgb_free(msg);
-		osmo_stream_cli_reconnect(cli);
+		stream_cli_reconnect(cli);
 		return 0;
 	}
 
@@ -353,12 +417,14 @@ static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 	OSMO_ASSERT(fd >= 0);
 
 	if (ret < 0) {
-		osmo_stream_cli_reconnect(cli);
+		LOGSCLI(cli, LOGL_ERROR, "connect failed (%d)\n", res);
+		stream_cli_reconnect(cli);
 		return;
 	}
 	ret = getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len);
 	if (ret >= 0 && error > 0) {
-		osmo_stream_cli_reconnect(cli);
+		LOGSCLI(cli, LOGL_ERROR, "connect so_error (%d)\n", error);
+		stream_cli_reconnect(cli);
 		return;
 	}
 
@@ -387,8 +453,11 @@ static void stream_cli_handle_connecting(struct osmo_stream_cli *cli, int res)
 	default:
 		break;
 	}
+	cli->in_cb_mask |= IN_CB_MASK_CONNECT_CB;
 	if (cli->connect_cb)
 		cli->connect_cb(cli);
+	cli->in_cb_mask &= ~IN_CB_MASK_CONNECT_CB;
+	free_delayed_if_needed(cli);
 }
 
 static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
@@ -402,11 +471,12 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 	case STREAM_CLI_STATE_CONNECTED:
 		if (what & OSMO_FD_READ) {
 			LOGSCLI(cli, LOGL_DEBUG, "connected read\n");
-			osmo_stream_cli_read(cli);
+			if (stream_cli_read(cli) == true)
+				break; /* cli (and hence ofd) freed, done. */
 		}
 		if (what & OSMO_FD_WRITE) {
 			LOGSCLI(cli, LOGL_DEBUG, "connected write\n");
-			osmo_stream_cli_write(cli);
+			stream_cli_write(cli);
 		}
 		break;
 	default:
@@ -443,6 +513,7 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 	cli->reconnect_timeout = 5;	/* default is 5 seconds. */
 	cli->segmentation_cb = NULL;
 	INIT_LLIST_HEAD(&cli->tx_queue);
+	cli->tx_queue_max_length = 1024; /* Default tx queue size, msgbs. */
 
 	cli->ma_pars.sctp.version = 0;
 
@@ -452,6 +523,7 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 static void stream_cli_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg)
 {
 	struct osmo_stream_cli *cli  = osmo_iofd_get_data(iofd);
+	bool freed;
 
 	switch (cli->state) {
 	case STREAM_CLI_STATE_CONNECTING:
@@ -463,21 +535,29 @@ static void stream_cli_iofd_read_cb(struct osmo_io_fd *iofd, int res, struct msg
 		case -EPIPE:
 		case -ECONNRESET:
 			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv (%d)\n", res);
-			osmo_stream_cli_reconnect(cli);
+			freed = stream_cli_reconnect(cli);
 			break;
 		case 0:
 			LOGSCLI(cli, LOGL_NOTICE, "connection closed with srv\n");
-			osmo_stream_cli_reconnect(cli);
+			freed = stream_cli_reconnect(cli);
 			break;
 		default:
 			LOGSCLI(cli, LOGL_DEBUG, "received %d bytes from srv\n", res);
+			freed = false;
 			break;
 		}
+		if (freed)
+			return; /* msg was also freed as part of the talloc tree. */
 		/* Notify user of new data or error: */
-		if (cli->iofd_read_cb)
-			cli->iofd_read_cb(cli, res, msg);
-		else
+		if (!cli->iofd_read_cb) {
 			msgb_free(msg);
+			return;
+		}
+		cli->in_cb_mask |= IN_CB_MASK_READ_CB;
+		cli->iofd_read_cb(cli, res, msg);
+		cli->in_cb_mask &= ~IN_CB_MASK_READ_CB;
+		OSMO_ASSERT(cli->in_cb_mask == 0);
+		free_delayed_if_needed(cli);
 		break;
 	default:
 		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
@@ -494,8 +574,8 @@ static void stream_cli_iofd_write_cb(struct osmo_io_fd *iofd, int res, struct ms
 		break;
 	case STREAM_CLI_STATE_CONNECTED:
 		if (msg && res <= 0) {
-			osmo_stream_cli_reconnect(cli);
 			LOGSCLI(cli, LOGL_ERROR, "received error %d in response to send\n", res);
+			stream_cli_reconnect(cli);
 		}
 		break;
 	default:
@@ -514,6 +594,7 @@ static const struct osmo_io_ops osmo_stream_cli_ioops = {
 static void stream_cli_iofd_recvmsg_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg, const struct msghdr *msgh)
 {
 	struct osmo_stream_cli *cli  = osmo_iofd_get_data(iofd);
+	bool freed;
 
 	res = stream_iofd_sctp_recvmsg_trailer(iofd, msg, res, msgh);
 
@@ -527,20 +608,28 @@ static void stream_cli_iofd_recvmsg_cb(struct osmo_io_fd *iofd, int res, struct
 		case -EPIPE:
 		case -ECONNRESET:
 			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv (%d)\n", res);
-			osmo_stream_cli_reconnect(cli);
+			freed = stream_cli_reconnect(cli);
 			break;
 		case 0:
 			LOGSCLI(cli, LOGL_NOTICE, "connection closed with srv\n");
-			osmo_stream_cli_reconnect(cli);
+			freed = stream_cli_reconnect(cli);
 			break;
 		default:
+			freed = false;
 			break;
 		}
+		if (freed)
+			return; /* msg was also freed as part of the talloc tree. */
 		/* Notify user of new data or error: */
-		if (cli->iofd_read_cb)
-			cli->iofd_read_cb(cli, res, msg);
-		else
+		if (!cli->iofd_read_cb) {
 			msgb_free(msg);
+			return;
+		}
+		cli->in_cb_mask |= IN_CB_MASK_READ_CB;
+		cli->iofd_read_cb(cli, res, msg);
+		cli->in_cb_mask &= ~IN_CB_MASK_READ_CB;
+		OSMO_ASSERT(cli->in_cb_mask == 0);
+		free_delayed_if_needed(cli);
 		break;
 	default:
 		osmo_panic("%s() called with unexpected state %d\n", __func__, cli->state);
@@ -778,6 +867,24 @@ void *osmo_stream_cli_get_data(struct osmo_stream_cli *cli)
 	return cli->data;
 }
 
+/*! Set the maximum length queue of the stream client.
+ *  \param[in] cli Stream Client to modify
+ *  \param[in] size maximum amount of msgbs which can be queued in the internal tx queue.
+ *  \returns 0 on success, negative on error.
+ *
+ *  The maximum length queue default value is 1024 msgbs. */
+int osmo_stream_cli_set_tx_queue_max_length(struct osmo_stream_cli *cli, unsigned int size)
+{
+	cli->tx_queue_max_length = size;
+
+	if (cli->iofd) /* Otherwise, this will be done in osmo_stream_cli_open() */
+		osmo_iofd_set_txqueue_max_length(cli->iofd, cli->tx_queue_max_length);
+
+	/* XXX: Here, in OSMO_STREAM_MODE_OSMO_FD mode we could check current
+	 * size of cli->tx_queue and shrink it from the front or back... */
+	return 0;
+}
+
 /*! Retrieve the stream client socket description.
  *  Calling this function will build a string that describes the socket in terms of its local/remote
  *  address/port.  The returned name is stored in a static buffer; it is hence not re-entrant or thread-safe.
@@ -855,10 +962,27 @@ osmo_stream_cli_set_read_cb2(struct osmo_stream_cli *cli,
  *  \param[in] cli Stream Client to destroy */
 void osmo_stream_cli_destroy(struct osmo_stream_cli *cli)
 {
-	osmo_stream_cli_close(cli);
+	if (!cli)
+		return;
+
+	LOGSCLI(cli, LOGL_DEBUG, "destroy()\n");
+	OSMO_ASSERT(!stream_cli_close(cli));
 	osmo_timer_del(&cli->timer);
 	msgb_queue_free(&cli->tx_queue);
-	talloc_free(cli);
+	cli->tx_queue_count = 0;
+	/* if we are in a user callback, delay freeing. */
+	if (cli->in_cb_mask != 0) {
+		LOGSCLI(cli, LOGL_DEBUG, "delay free() in_cb_mask=0x%02x\n", cli->in_cb_mask);
+		cli->delay_free = true;
+		/* Move ptr to avoid double free if parent ctx of cli is freed
+		 * meanwhile (eg. during user callback after calling
+		 * osmo_stream_client_destroy() and before returning from user
+		 * callback. */
+		talloc_steal(OTC_GLOBAL, cli);
+	} else {
+		LOGSCLI(cli, LOGL_DEBUG, "free(destroy)\n");
+		talloc_free(cli);
+	}
 }
 
 /*! DEPRECATED: use osmo_stream_cli_set_reconnect_timeout() or osmo_stream_cli_reconnect() instead!
@@ -872,8 +996,10 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 	int ret;
 
 	/* we are reconfiguring this socket, close existing first. */
-	if ((cli->flags & OSMO_STREAM_CLI_F_RECONF) && cli->ofd.fd >= 0)
-		osmo_stream_cli_close(cli);
+	if ((cli->flags & OSMO_STREAM_CLI_F_RECONF) && cli->ofd.fd >= 0) {
+		if (stream_cli_close(cli) == true)
+			return -ENAVAIL; /* freed */
+	}
 
 	cli->flags &= ~OSMO_STREAM_CLI_F_RECONF;
 
@@ -883,7 +1009,8 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 		ret = osmo_sock_init2_multiaddr2(AF_UNSPEC, SOCK_STREAM, cli->proto,
 						(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
 						(const char **)cli->addr, cli->addrcnt, cli->port,
-						OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK,
+						OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_BIND | OSMO_SOCK_F_NONBLOCK |
+						OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio),
 						&cli->ma_pars);
 		break;
 #endif
@@ -891,12 +1018,14 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect)
 		ret = osmo_sock_init2(AF_UNSPEC, SOCK_STREAM, cli->proto,
 				      cli->local_addr[0], cli->local_port,
 				      cli->addr[0], cli->port,
-				      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+				      OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_BIND | OSMO_SOCK_F_NONBLOCK |
+				      OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio));
 	}
 
 	if (ret < 0) {
+		LOGSCLI(cli, LOGL_ERROR, "connect: socket creation error (%d)\n", ret);
 		if (reconnect)
-			osmo_stream_cli_reconnect(cli);
+			stream_cli_reconnect(cli);
 		return ret;
 	}
 	osmo_fd_setup(&cli->ofd, ret, OSMO_FD_READ | OSMO_FD_WRITE, osmo_stream_cli_fd_cb, cli, 0);
@@ -947,6 +1076,69 @@ void osmo_stream_cli_set_nodelay(struct osmo_stream_cli *cli, bool nodelay)
 			nodelay, errno);
 }
 
+/*! Set the priority value of the stream socket.
+ *  Setting this  will automatically set the socket priority
+ *  option on any socket established via \ref osmo_stream_cli_open
+ *  or any re-connect. This can be set either before or after opening the
+ *  socket.
+ *  \param[in] cli Stream client whose sockets are to be configured
+ *  \param[in] sk_prio priority value. Values outside 0..6 require CAP_NET_ADMIN.
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_cli_set_priority(struct osmo_stream_cli *cli, int sk_prio)
+{
+	int rc;
+	int fd;
+
+	if (cli->sk_prio == sk_prio)
+		return 0; /* No change needed */
+
+	cli->sk_prio = sk_prio;
+
+	if (!stream_cli_is_opened(cli))
+		return 0; /* Config will be applied upon open() time */
+
+	if ((fd = osmo_stream_cli_get_fd(cli)) < 0) { /* Shouldn't happen... */
+		LOGSCLI(cli, LOGL_ERROR, "set_priority(%d): failed obtaining socket\n", cli->sk_prio);
+		return -EBADFD;
+	}
+	if ((rc = osmo_sock_set_priority(fd, cli->sk_prio)) < 0)
+		LOGSCLI(cli, LOGL_ERROR, "set_priority(%d): failed setsockopt err=%d\n",
+			cli->sk_prio, errno);
+	return rc;
+}
+
+/*! Set the DSCP (differentiated services code point) of the stream socket.
+ *  Setting this  will automatically set the IP DSCP option on any socket established
+ *  via \ref osmo_stream_cli_open or any re-connect. This can be set either before or
+ *  after opening the socket.
+ *  \param[in] cli Stream client whose sockets are to be configured
+ *  \param[in] ip_dscp DSCP value. Value range 0..63.
+ *  \return negative on error, 0 on success
+ */
+int osmo_stream_cli_set_ip_dscp(struct osmo_stream_cli *cli, uint8_t ip_dscp)
+{
+	int rc;
+	int fd;
+
+	if (cli->ip_dscp == ip_dscp)
+		return 0; /* No change needed */
+
+	cli->ip_dscp = ip_dscp;
+
+	if (!stream_cli_is_opened(cli))
+		return 0; /* Config will be applied upon open() time */
+
+	if ((fd = osmo_stream_cli_get_fd(cli)) < 0) { /* Shouldn't happen... */
+		LOGSCLI(cli, LOGL_ERROR, "set_ip_dscp(%u): failed obtaining socket\n", cli->ip_dscp);
+		return -EBADFD;
+	}
+	if ((rc = osmo_sock_set_dscp(fd, cli->ip_dscp)) < 0)
+		LOGSCLI(cli, LOGL_ERROR, "set_ip_dscp(%u): failed setsockopt err=%d\n",
+			cli->ip_dscp, errno);
+	return rc;
+}
+
 /*! Open connection of an Osmocom stream client.
  *  This will initiate an non-blocking outbound connect to the configured destination (server) address.
  *  By default the client will automatically attempt to reconnect after default timeout.
@@ -960,8 +1152,10 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 	unsigned int local_addrcnt;
 
 	/* we are reconfiguring this socket, close existing first. */
-	if ((cli->flags & OSMO_STREAM_CLI_F_RECONF) && osmo_stream_cli_get_fd(cli) >= 0)
-		osmo_stream_cli_close(cli);
+	if ((cli->flags & OSMO_STREAM_CLI_F_RECONF) && osmo_stream_cli_get_fd(cli) >= 0) {
+		if (stream_cli_close(cli) == true)
+			return -ENAVAIL; /* freed */
+	}
 
 	cli->flags &= ~OSMO_STREAM_CLI_F_RECONF;
 
@@ -976,7 +1170,8 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 #ifdef HAVE_LIBSCTP
 		case IPPROTO_SCTP:
 			local_addrcnt = cli->local_addrcnt;
-			flags = OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_NONBLOCK;
+			flags = OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_NONBLOCK |
+				OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio);
 			if (cli->local_addrcnt > 0 || cli->local_port > 0) { /* explicit bind required? */
 				flags |= OSMO_SOCK_F_BIND;
 				/* If no local addr configured, use local_addr[0]=NULL by default when creating the socket. */
@@ -993,7 +1188,8 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 			ret = osmo_sock_init2(cli->sk_domain, cli->sk_type, cli->proto,
 					      cli->local_addr[0], cli->local_port,
 					      cli->addr[0], cli->port,
-					      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+					      OSMO_SOCK_F_CONNECT | OSMO_SOCK_F_BIND | OSMO_SOCK_F_NONBLOCK |
+					      OSMO_SOCK_F_DSCP(cli->ip_dscp) | OSMO_SOCK_F_PRIO(cli->sk_prio));
 		}
 		break;
 	default:
@@ -1001,7 +1197,8 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 	}
 
 	if (ret < 0) {
-		osmo_stream_cli_reconnect(cli);
+		LOGSCLI(cli, LOGL_ERROR, "connect: socket creation error (%d)\n", ret);
+		stream_cli_reconnect(cli);
 		return ret;
 	}
 
@@ -1021,7 +1218,7 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
 		/* Be sure that previous osmo_io instance is freed before creating a new one. */
-		osmo_stream_cli_close_iofd(cli);
+		stream_cli_close_iofd(cli);
 #ifdef HAVE_LIBSCTP
 		if (cli->proto == IPPROTO_SCTP) {
 			cli->iofd = osmo_iofd_setup(cli, fd, cli->name, OSMO_IO_FD_MODE_RECVMSG_SENDMSG,
@@ -1038,8 +1235,8 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 		if (!cli->iofd)
 			goto error_close_socket;
 
+		osmo_iofd_set_txqueue_max_length(cli->iofd, cli->tx_queue_max_length);
 		osmo_iofd_notify_connected(cli->iofd);
-
 		configure_cli_segmentation_cb(cli, cli->segmentation_cb);
 
 		if (osmo_iofd_register(cli->iofd, fd) < 0)
@@ -1087,7 +1284,12 @@ void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg)
 
 	switch (cli->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
-		msgb_enqueue(&cli->tx_queue, msg);
+		if (cli->tx_queue_count >= cli->tx_queue_max_length) {
+			LOGSCLI(cli, LOGL_ERROR, "send: tx queue full, dropping msg!\n");
+			msgb_free(msg);
+			return;
+		}
+		msgb_enqueue_count(&cli->tx_queue, msg, &cli->tx_queue_count);
 		osmo_fd_write_enable(&cli->ofd);
 		break;
 	case OSMO_STREAM_MODE_OSMO_IO:
@@ -1159,11 +1361,13 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 	}
 
 	if (ret < 0) {
-		if (ret == -EAGAIN)
+		if (ret == -EAGAIN) /* Received MSG_NOTIFICATION from stream_sctp_recvmsg_wrapper() */
 			return ret;
 		if (errno == EPIPE || errno == ECONNRESET)
-			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv\n");
-		osmo_stream_cli_reconnect(cli);
+			LOGSCLI(cli, LOGL_ERROR, "lost connection with srv (%d)\n", errno);
+		else
+			LOGSCLI(cli, LOGL_ERROR, "recv failed (%d)\n", errno);
+		stream_cli_reconnect(cli);
 		return ret;
 	} else if (ret == 0) {
 		LOGSCLI(cli, LOGL_ERROR, "connection closed with srv\n");
@@ -1182,6 +1386,7 @@ void osmo_stream_cli_clear_tx_queue(struct osmo_stream_cli *cli)
 	switch (cli->mode) {
 	case OSMO_STREAM_MODE_OSMO_FD:
 		msgb_queue_free(&cli->tx_queue);
+		cli->tx_queue_count = 0;
 		/* If in state 'connecting', keep WRITE flag up to receive
 		* socket connection signal and then transition to STATE_CONNECTED: */
 		if (cli->state == STREAM_CLI_STATE_CONNECTED)
